import os
import cv2
import numpy as np
import onnxruntime as nxrun
from .face_detector import FaceDetector

class ImageFeatureGenerator:
    def __init__(self, ident_onnx_path, quality_onnx_path, detect_onnx_path, face_id_tflite_path):
        self._ident_onnx_path = ident_onnx_path
        self._quality_onnx_path =  quality_onnx_path
        self._face_id_tflite_path = face_id_tflite_path
        self._face_detector = FaceDetector(detect_onnx_path)
    
    # get align face
    def _detect_face(self, image):
        # print(image.shape)
        _, align_face,_,_ = self._face_detector.detect_and_align(image)
        if align_face.shape[2] != 3:
            return []
        return align_face
    
    # generete feature from align face
    def generate_feature(self, img):
        sess = nxrun.InferenceSession(self._ident_onnx_path)
        
        img = cv2.resize(img, (112,112))
        img = img[:, :, ::-1]
        
        input_data = np.array([img], dtype=np.float32)
        input_data = np.transpose(input_data, (0,3,1,2))
        input_name = sess.get_inputs()[0].name
        feature = sess.run(None, {input_name: input_data})[0][0]
        
        del sess
        
        return feature

    def generate_feature_128(self, img):
        import tensorflow as tf
        interpreter = tf.lite.Interpreter(model_path=self._face_id_tflite_path)

        interpreter.allocate_tensors()
        input_details = interpreter.get_input_details()
        output_details = interpreter.get_output_details()

        img = cv2.resize(img, (112,112))
        img = img[:, :, ::-1]
        
        input_data = np.array([img], dtype=np.float32)
        # input_data = np.transpose(input_data, (0,3,1,2))

        interpreter.set_tensor(input_details[0]['index'], input_data)
        interpreter.invoke()
        output_data = interpreter.get_tensor(output_details[0]['index'])
        return output_data[0]

    def get_align_face(self, image_path):
        img = cv2.imread(image_path)
        align_face = self._detect_face(img)
        return align_face

    def get_quality(self, img):
        labels = ["bad", "good", "mask", "very_good"]
        sess = nxrun.InferenceSession(self._quality_onnx_path)
        img = cv2.resize(img, (112,112))
        img = img / 127.5
        img = img - 1
        img = img[:, :, ::-1]
        input_data = np.array([img], dtype=np.float32)
        input_data = np.transpose(input_data, (0,3,1,2))
        input_name = sess.get_inputs()[0].name
        result = sess.run(None, {input_name: input_data})[0][0]
        predict = np.argmax(result)
        del sess
        return predict, labels[predict]

    def get_face_quality(self, image_path):
        img = self.get_align_face(image_path)
        if img == []:
            return -1, -1
        idx, label = self.get_quality(img)
        return idx, label

    # case: receive base64 or array image
    def get_feature(self, img):
        pass
    
    # case: receive image_path from somewhere such as minio
    def get_feature_from_path(self, image_path):
        img = cv2.imread(image_path)
        align_face = self._detect_face(img)
        if align_face == []:
            return -1
        feature = self.generate_feature(align_face)
        return feature
    
    def get_feature_128_from_path(self, image_path):
        img = cv2.imread(image_path)
        align_face = self._detect_face(img)
        if align_face == []:
            return -1
        feature = self.generate_feature_128(align_face)
        return feature

# if __name__ == "__main__":
#     image_path = "test_data/69264356101573563301609675830640888568610816n-15696603325551387270557.jpg"
#     # ident_onnx_path = 'facefeaturelib/onnx/identification_mask.onnx'
#     ident_onnx_path = 'onnx/face_id_mobile.onnx'
#     face_id_tflite_path = 'tflite/tf_arcface_100_v1.tflite'
#     quality_onnx_path = 'onnx/face_quality_42.onnx'
#     detect_onnx_path = 'onnx/scrfd_2.5g_bnkps_shape640x640.onnx'
#     generator = ImageFeatureGenerator( ident_onnx_path, quality_onnx_path, detect_onnx_path, face_id_tflite_path)
#     feature = generator.get_feature_128_from_path(image_path)
#     # feature = generator.get_feature_from_path(image_path)
    # print(feature.shape)
    # idx, label = generator.get_face_quality(image_path)
    # print(idx, label)


import cv2
import numpy as np
import onnxruntime as nxrun

class SCRFD:
    def __init__(self, detect_onnx_path):
        self.center_cache = {}
        self.conf_thresh = 0.3
        self.nms_thresh = 0.45
        self._detect_onnx_path = detect_onnx_path
        self.input_size = (640, 640)
        self.fmc = 3
        self._feat_stride_fpn = [8, 16, 32]
        self._num_anchors = 2
        self.use_kps = True

    def _forward(self, img, thresh):
        scores_list = []
        bboxes_list = []
        kpss_list = []

        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB).astype(np.float32)
        img = (img - np.array([127.5, 127.5, 127.5], dtype=np.float32)) / 128.0
        img = np.transpose(img, (2, 0, 1))
        img = np.expand_dims(img, axis=0)
        sess = nxrun.InferenceSession(self._detect_onnx_path)
        input_name = sess.get_inputs()[0].name
        net_outs = sess.run(None, {input_name: img})
        orders = [0,3,6,1,4,7,2,5,8]
        new_net_outs = []
        for order in orders:
            new_net_outs.append(net_outs[order])
        net_outs = new_net_outs
        del sess
        input_height = img.shape[2]
        input_width = img.shape[3]
        fmc = self.fmc
        for idx, stride in enumerate(self._feat_stride_fpn):
            scores = net_outs[idx*fmc]
            bbox_preds = net_outs[idx*fmc+1]
            bbox_preds = bbox_preds * stride
            if self.use_kps:
                kps_preds = net_outs[idx*fmc+2] * stride
            height = input_height // stride
            width = input_width // stride
            key = (height, width, stride)
            if key in self.center_cache:
                anchor_centers = self.center_cache[key]
            else:
                # solution-3:
                anchor_centers = np.stack(
                    np.mgrid[:height, :width][::-1], axis=-1).astype(np.float32)
                anchor_centers = (anchor_centers * stride).reshape((-1, 2))
                if self._num_anchors > 1:
                    anchor_centers = np.stack([anchor_centers]*self._num_anchors, axis=1).reshape((-1, 2))
                if len(self.center_cache) < 100:
                    self.center_cache[key] = anchor_centers

            pos_inds = np.where(scores >= thresh)[0]
            bboxes = SCRFD._distance2bbox(anchor_centers, bbox_preds)
            pos_scores = scores[pos_inds]
            pos_bboxes = bboxes[pos_inds]
            scores_list.append(pos_scores)
            bboxes_list.append(pos_bboxes)
            if self.use_kps:
                kpss = SCRFD._distance2kps(anchor_centers, kps_preds)
                kpss = kpss.reshape((kpss.shape[0], -1, 2))
                pos_kpss = kpss[pos_inds]
                kpss_list.append(pos_kpss)
        return scores_list, bboxes_list, kpss_list

    def detect(self, img):
        im_ratio = float(img.shape[0]) / img.shape[1]
        model_ratio = float(self.input_size[1]) / self.input_size[0]
        if im_ratio > model_ratio:
            new_height = self.input_size[1]
            new_width = int(new_height / im_ratio)
        else:
            new_width = self.input_size[0]
            new_height = int(new_width * im_ratio)
        det_scale = float(new_height) / img.shape[0]
        resized_img = cv2.resize(img, (new_width, new_height))
        det_img = np.zeros((self.input_size[1], self.input_size[0], 3), dtype=np.uint8)
        det_img[:new_height, :new_width, :] = resized_img

        scores_list, bboxes_list, kpss_list = self._forward(det_img, self.conf_thresh)

        scores = np.vstack(scores_list)
        scores_ravel = scores.ravel()
        order = scores_ravel.argsort()[::-1]
        bboxes = np.vstack(bboxes_list) / det_scale
        if self.use_kps:
            kpss = np.vstack(kpss_list) / det_scale
        pre_det = np.hstack((bboxes, scores)).astype(np.float32, copy=False)
        pre_det = pre_det[order, :]
        keep = self._nms(pre_det)
        det = pre_det[keep, :]
        if self.use_kps:
            kpss = kpss[order, :, :]
            kpss = kpss[keep, :, :]
        else:
            kpss = None
        return det, kpss

    def _nms(self, dets):
        thresh = self.nms_thresh
        x1 = dets[:, 0]
        y1 = dets[:, 1]
        x2 = dets[:, 2]
        y2 = dets[:, 3]
        scores = dets[:, 4]

        areas = (x2 - x1 + 1) * (y2 - y1 + 1)
        order = scores.argsort()[::-1]

        keep = []
        while order.size > 0:
            i = order[0]
            keep.append(i)
            xx1 = np.maximum(x1[i], x1[order[1:]])
            yy1 = np.maximum(y1[i], y1[order[1:]])
            xx2 = np.minimum(x2[i], x2[order[1:]])
            yy2 = np.minimum(y2[i], y2[order[1:]])

            w = np.maximum(0.0, xx2 - xx1 + 1)
            h = np.maximum(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (areas[i] + areas[order[1:]] - inter)

            inds = np.where(ovr <= thresh)[0]
            order = order[inds + 1]

        return keep

    @staticmethod
    def _distance2bbox(points, distance, max_shape=None):
        """Decode distance prediction to bounding box.

        Args:
            points (Tensor): Shape (n, 2), [x, y].
            distance (Tensor): Distance from the given point to 4
                boundaries (left, top, right, bottom).
            max_shape (tuple): Shape of the image.

        Returns:
            Tensor: Decoded bboxes.
        """
        x1 = points[:, 0] - distance[:, 0]
        y1 = points[:, 1] - distance[:, 1]
        x2 = points[:, 0] + distance[:, 2]
        y2 = points[:, 1] + distance[:, 3]
        if max_shape is not None:
            x1 = x1.clamp(min=0, max=max_shape[1])
            y1 = y1.clamp(min=0, max=max_shape[0])
            x2 = x2.clamp(min=0, max=max_shape[1])
            y2 = y2.clamp(min=0, max=max_shape[0])
        return np.stack([x1, y1, x2, y2], axis=-1)

    @staticmethod
    def _distance2kps(points, distance, max_shape=None):
        """Decode distance prediction to bounding box.

        Args:
            points (Tensor): Shape (n, 2), [x, y].
            distance (Tensor): Distance from the given point to 4
                boundaries (left, top, right, bottom).
            max_shape (tuple): Shape of the image.

        Returns:
            Tensor: Decoded bboxes.
        """
        preds = []
        for i in range(0, distance.shape[1], 2):
            px = points[:, i % 2] + distance[:, i]
            py = points[:, i % 2+1] + distance[:, i+1]
            if max_shape is not None:
                px = px.clamp(min=0, max=max_shape[1])
                py = py.clamp(min=0, max=max_shape[0])
            preds.append(px)
            preds.append(py)
        return np.stack(preds, axis=-1)




import cv2
import numpy as np
from skimage import transform as trans

from .scrfd import SCRFD

class FaceDetector:
    def __init__(self, detect_onnx_path):
        self.face_detection = SCRFD(detect_onnx_path)
        self.img_thumb_width = 320 #640#480
        self.img_thumb_height = 400 #640#270
        self.detect_img_size = 112
        self.batch_size = 4

        self.margin_time = 1.25#0.7

        self.detect_sensitive_low = 0.5
        self.detect_sensitive_norm = 0.65
        self.detect_sensitive_high = 0.8
    def detect_face(self, img_file_path):

        img1 = cv2.imread(img_file_path)
        results2 = self.face_detection.detect(img1)
        detections = results2[0]
        landmarks = results2[1]

        scale_width = 1
        scale_height = 1

        outbboxes = None
        outlandmarks = None
        max_size = 0
        for i, result in enumerate(detections):
            result[0] = result[0] * scale_width
            result[2] = result[2] * scale_width
            result[1] = result[1] * scale_height
            result[3] = result[3] * scale_height
            result = [round(x) for x in result]
            landmark = np.asarray(landmarks[i])

            landmark[:, 0] = landmark[:, 0] * scale_width
            landmark[:, 1] = landmark[:, 1] * scale_height
            tem_size = ((result[2] - result[0]) + (result[3] - result[1])) / 2
            if tem_size > max_size:
                max_size = tem_size
                outbboxes = result.copy()
                outlandmarks = landmark.copy()
        return outbboxes, np.asarray(outlandmarks).astype(np.int)
    
    def get_face_img(self, img1):
        results2 = self.face_detection.detect(img1)
        detections = results2[0]

        scale_width = 1
        scale_height = 1

        outbboxes = None
        max_size = 0
        for _, result in enumerate(detections):
            result[0] = result[0] * scale_width
            result[2] = result[2] * scale_width
            result[1] = result[1] * scale_height
            result[3] = result[3] * scale_height
            result = [round(x) for x in result]

            tem_size = ((result[2] - result[0]) + (result[3] - result[1])) / 2
            if tem_size > max_size:
                max_size = tem_size
                outbboxes = result.copy()

        img_dis = None
        if np.shape(outbboxes) !=():
            img_dis = self.get_thumb(img1, outbboxes)
        return img_dis
    def detect_face_img(self, img1):
        img1 = np.transpose(img1, (1 ,2, 0))
        img_shape = np.shape(img1)
        y_cen = img_shape[0]/2
        x_cen = img_shape[1]/2
        results2 = self.face_detection.detect(img1)
        detections = results2[0]
        landmarks = results2[1]

        scale_width = 1
        scale_height = 1

        outbboxes = None
        outlandmarks = None
        dis_cen = 10000
        for i, result in enumerate(detections):
            result[0] = result[0] * scale_width
            result[2] = result[2] * scale_width
            result[1] = result[1] * scale_height
            result[3] = result[3] * scale_height
            result = [round(x) for x in result]
            landmark = np.asarray(landmarks[i])

            landmark[:, 0] = landmark[:, 0] * scale_width
            landmark[:, 1] = landmark[:, 1] * scale_height
            x_cen_tem = (result[2] + result[0])/2
            y_cen_tem = (result[3] + result[1])/2
            tem_size = np.sqrt((x_cen_tem -x_cen)*(x_cen_tem -x_cen) + (y_cen_tem - y_cen)*(y_cen_tem - y_cen))
            if tem_size < dis_cen:
                dis_cen = tem_size
                outbboxes = result.copy()
                outlandmarks = landmark.copy()
        if np.shape(outlandmarks) !=():
            outlandmarks =np.asarray(outlandmarks).astype(np.int)
        return outbboxes, outlandmarks


    def align_face(self, img, bbox, landmark):
        warp_img = self.preprocess_detect(img, bbox=bbox, landmark=landmark,
                                          img_size=str(self.detect_img_size)+','+str(self.detect_img_size))
        return warp_img

    def get_norm_from_thumb(self, img, bbox, landmark_points):
        img = np.transpose(img, (1, 2, 0))
        img_align = self.align_face(img, bbox, landmark_points)
        return img_align

    def get_thumb(self, img, bbox):
        (x, y, x2, y2, score) = bbox
        img_size = np.asarray(img.shape)[0:2]
        margin_x = int(self.margin_time * (x2 - x) / 2)
        margin_y = int(self.margin_time * (y2 - y) / 2)
        x_dis = int(np.maximum(x - margin_x, 0))
        y_dis = int(np.maximum(y - margin_y, 0))
        x_dis2 = int(np.minimum(x2 + margin_x, img_size[1]))
        y_dis2 = int(np.minimum(y2 + margin_y, img_size[0]))

        img_thumb = img[y_dis:y_dis2, x_dis:x_dis2, :]
        img_thumb = cv2.resize(img_thumb, (self.img_thumb_width, self.img_thumb_height))
        return img_thumb
    ## From origin full images
    def detect_and_align(self, img):
        ## Detect thumb imgs
        shape_img = np.shape(img)
        if shape_img[0] == self.img_thumb_height and shape_img[1] == self.img_thumb_width:
            img_thumb_tem = img.copy()
        else:
            img_thumb_tem = self.get_face_img(img)
        if np.shape(img_thumb_tem) ==():
            img_thumb_tem = cv2.resize(img, (self.img_thumb_width, self.img_thumb_height))
        ## Detect landmark, align
        bbox, img_align, score = self.get_align_from_thumb(img_thumb_tem)
        return bbox, img_align, img_thumb_tem, score
    ## From thumb images
    def get_align_from_thumb(self, img_thumb_tem):
        bbox = None
        score = 0.0
        img_thumb_tem = np.transpose(img_thumb_tem, (2, 0, 1))
        info, landmark = self.detect_face_img(img_thumb_tem)
        if np.shape(landmark) !=():
            bbox = info[:4]
            score = info[4]
            img_align = self.get_norm_from_thumb(img_thumb_tem, bbox, landmark)
        else:
            img_align = cv2.resize(img_thumb_tem, (self.detect_img_size, self.detect_img_size))
        return bbox, img_align, float(score)

    def preprocess_detect(self, img, landmark=None, bbox=None, **kwargs):
        image_size = [112, 112]
        M = None
        if landmark is not None:
            assert len(image_size) == 2
            src = np.array([
                [30.2946, 51.6963],
                [65.5318, 51.5014],
                [48.0252, 71.7366],
                [33.5493, 92.3655],
                [62.7299, 92.2041]], dtype=np.float32)
            if image_size[1] == 112:
                src[:, 0] += 8.0
            dst = landmark.astype(np.float32)

            tform = trans.SimilarityTransform()
            tform.estimate(dst, src)
            M = tform.params[0:2, :]

        if M is None:
            if bbox is None:  # use center crop
                det = np.zeros(4, dtype=np.int32)
                det[0] = int(img.shape[1] * 0.0625)
                det[1] = int(img.shape[0] * 0.0625)
                det[2] = img.shape[1] - det[0]
                det[3] = img.shape[0] - det[1]
            else:
                det = bbox
            margin = kwargs.get('margin', 44)
            bb = np.zeros(4, dtype=np.int32)
            bb[0] = np.maximum(det[0] - margin / 2, 0)
            bb[1] = np.maximum(det[1] - margin / 2, 0)
            bb[2] = np.minimum(det[2] + margin / 2, img.shape[1])
            bb[3] = np.minimum(det[3] + margin / 2, img.shape[0])
            ret = img[bb[1]:bb[3], bb[0]:bb[2], :]
            if len(image_size) > 0:
                ret = cv2.resize(ret, (image_size[1], image_size[0]))
            return ret
        else:  # do align using landmark
            assert len(image_size) == 2
            warped = cv2.warpAffine(img, M, (image_size[1], image_size[0]), borderValue=0.0)

            return warped



from setuptools import find_packages, setup
setup(
    name='facefeaturelib',
    packages=find_packages(include=['facefeaturelib']),
    version='0.1.0',
    description='image feature library',
    author='hoangvd3',
    license='MIT',
    install_requires=[],
)